/* $Id: joblist.cc,v 1.19 2002/02/13 00:36:16 richard Exp $ -*- C++ -*-
  __   _
  |_) /|  Copyright (C) 2001-2002 Richard Atterer
  | \/|  <richard@atterer.net>
   '` 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2. See
  the file COPYING for details.

  Interface to the GtkCList of running jobs (downloads etc), window.jobs

*/

#include <iomanip>
#include <iostream>
namespace std { }
using namespace std;
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <config.h>
#if WINDOWS
#  include <gdkprivate.h>
#endif

#include <gtk-error.hh>
#include <gtk-support.hh>
#include <joblist.hh>
#include <string.hh>
//______________________________________________________________________

JobVector GUI::jobVector;

gchar* JobVector::noText[] = { 0, 0 }; // For init'ing CList row
vector<GdkPixmap*> JobVector::progressGfx;
//______________________________________________________________________

Job::~Job() {
  if (needTicks()) jobVector()->unregisterTicks();
}
//______________________________________________________________________

void Job::waitTick() {
  if (--waitCountdown == 0) callRegularly(waitDestination);
# if DEBUG_JOBLIST
    cerr << "waitCountdown=" << waitCountdown << endl;
# endif
}
//______________________________________________________________________

namespace {

  /* Returns true if last characters of uri are the same as ext.
     Cannot use string::compare because of incompatibility between
     GCC2.95/3.0 */
  inline bool compareEnd(const string& uri, const char* ext) {
    static const unsigned extLen = strlen(ext);
    if (uri.size() < extLen) return false;
    unsigned pos = uri.size() - extLen;
    for (unsigned i = 0; i < extLen; ++i)
      if (ext[i] != uri[pos + i]) return false;
    return true;
  }

}
//________________________________________

// Return a pixmap appropriate for the URI. Loads the gfx from file on demand
GtkPixmap* Job::pixmapForURI(const string& uri) {
  enum { DEFAULT, deb, jigdo, iso, LAST };
  static GtkPixmap* icon[LAST]; // Initialized to 0

  struct Mapping { unsigned extLen; const char* ext; int icon; };
  const Mapping map[] = {
    { 4, ".deb",      deb },
    { 4, ".dsc",      deb },
    { 6, ".jigdo",    jigdo },
    { 9, ".template", jigdo },
    { 4, ".iso",      iso },
    { 4, ".raw",      iso }
  };
  //____________________

  // Search through the map entries
  int type = DEFAULT;
  string::const_iterator end = uri.end();
  for (size_t m = 0; m < sizeof(map) / sizeof(Mapping); ++m) {
    // extLen field for each entry must be length of ext string
    const char* ext = map[m].ext;
    Paranoid(map[m].extLen == strlen(ext));
    if (map[m].extLen > uri.size()) continue;
    string::const_iterator i = end - map[m].extLen;
    while (*ext != '\0' && *i == *ext) { ++i; ++ext; }
    if (*ext == '\0') { type = map[m].icon; break; }
  }
  GtkPixmap* result = icon[type];
  if (result == 0) {
    const char* filename = 0;
    switch (type) {
      case deb:     filename = "job-deb.xpm"; break;
      case jigdo:   filename = "job-jigdo.xpm"; break;
      case iso:     filename = "job-iso.xpm"; break;
      case DEFAULT: filename = "job-file.xpm"; break;
    }
#   if DEBUG
    //cerr << "Loading `" << filename << '\'' << endl;
#   endif
    result = GTK_PIXMAP(create_pixmap(GUI::window.window, filename));
    icon[type] = result;
  }
  return result;
}
//======================================================================

JobVector::~JobVector() {
  /* Don't do gtk_clist_clear() - the CList is auto-deleted because it
     is the child of a deleted window. */

  // Delete Jobs
  size_type rows = size();
  for (size_type row = 0; row < rows; ++row) delete get(row);
}
//______________________________________________________________________

/* Create progress pixmaps from the big "progress051.xpm" file. The
   reason why we have to resort to pixmaps and not use a
   GtkProgressBar is that GtkCList does not appear to allow you to
   insert an arbitrary widget into a CList, just pixmaps and text.
   Grrr.
   Called from gtk-gui.cc */
void JobVector::initAfterGtk() {
  Assert(GUI::jobVector.list()->columns == 2);
  Assert(GUI::jobVector.empty());
  Paranoid(sizeof(noText) == NR_OF_COLUMNS * sizeof(gchar*));

  GtkWidget* progress = create_pixmap(GUI::window.window, PROGRESS_XPM);
  GdkPixmap* progressImage;
  gint width = -1, height = -1, depth = -1;

  gtk_pixmap_get(GTK_PIXMAP(progress), &progressImage, NULL);
# if WINDOWS
  // This is probably the wrong way to get the width/height...
  width = ((GdkDrawablePrivate*)progressImage)->width;
  height = ((GdkDrawablePrivate*)progressImage)->height;
  depth = 24;
# else
  /* It is not clearly documented whether casting from GdkPixmap* to
     GdkWindow* in this way is OK... :-/ */
  gdk_window_get_geometry((GdkWindow*)progressImage, 0, 0,
                          &width, &height, &depth);
  Assert(width != -1 && height != -1 && depth != -1);
# endif
// # if DEBUG
//   cerr << PROGRESS_XPM << ": " << width << " x " << height << " x "
//        << depth << endl;
// # endif
  GdkGC* gc = gdk_gc_new((GdkWindow*)progressImage);
  Paranoid(gc != 0);
  //____________________

  /* Split up progressImage into the many smaller pixmaps for the
     graphical progress bar. These pixmaps will never be de-allocated
     again. */
  Paranoid(progressGfx.empty());
  progressGfx.reserve(PROGRESS_SUBDIV);
  if (width == 1 && height == 1) {
    // progressGfx not found - fill array with dummy pixmaps
    for (int i = 0; i < PROGRESS_SUBDIV; ++i)
      progressGfx.push_back(progressImage);
  } else {
    int sheight = height / PROGRESS_SUBDIV; // Height of small pixmap
    for (int i = 0; i < height; i += sheight) {
      // Create new pixmap and add to list
      GdkPixmap* pic = gdk_pixmap_new(NULL, width, sheight, depth);
      Paranoid(pic != 0);
      progressGfx.push_back(pic);
      gdk_draw_pixmap((GdkDrawable*)pic, gc, progressImage,
                      0, i, 0, 0, width, sheight);
    }
  }
  //____________________

  gdk_gc_destroy(gc);
  gtk_widget_destroy(progress);
}
//______________________________________________________________________

GdkPixmap* JobVector::progress(uint64 curAmount, uint64 totalAmount) {
  if (totalAmount == 0) return progressGfx[0];
  unsigned nr = PROGRESS_SUBDIV * curAmount / totalAmount;
  if (nr >= progressGfx.size()) nr = progressGfx.size() - 1;
  return progressGfx[nr];
}
//______________________________________________________________________

// void JobVector::deleteJob(void* j) {
//   // cerr << "delete Job "<<j<<endl;
//   delete static_cast<Job*>(j);
// }
//______________________________________________________________________

void JobVector::erase(size_type n) {
  // If we were not already collapsing, ensure timeout fnc is registered now
  if (entryCount() == size()) {
    registerTicks();
    collapseCountdown = COLLAPSE_INTERVAL / TICK_INTERVAL;
  }

  // Delete entry
  Job* j = get(n);
  if (j) {
    Paranoid(entryCountVal > 0);
    --entryCountVal;
    delete j; // Delete!
  }
  makeRowBlank(n);
}
//______________________________________________________________________

namespace {

  void copyCell(GtkCList* clist, gint srcRow, gint destRow, gint column) {
    gchar* text;
    guint8 spacing;
    GdkPixmap* pixmap;
    GdkBitmap* mask;
    switch (gtk_clist_get_cell_type(clist, srcRow, column)) {
    case GTK_CELL_EMPTY:
      gtk_clist_set_text(clist, destRow, column, "");
      break;
    case GTK_CELL_TEXT:
      gtk_clist_get_text(clist, srcRow, column, &text);
      gtk_clist_set_text(clist, destRow, column, text);
      break;
    case GTK_CELL_PIXMAP:
      gtk_clist_get_pixmap(clist, srcRow, column, &pixmap, &mask);
      gtk_clist_set_pixmap(clist, destRow, column, pixmap, mask);
      break;
    case GTK_CELL_PIXTEXT:
      gtk_clist_get_pixtext(clist, srcRow, column, &text, &spacing,
                            &pixmap, &mask);
      gtk_clist_set_pixtext(clist, destRow, column, text, spacing,
                            pixmap, mask);
      break;
    case GTK_CELL_WIDGET:
    default:
      Assert("unsupported");
    }
  }

}
//______________________________________________________________________

gint JobVector::timeoutCallback(gpointer jobVector) {
  JobVector* self = static_cast<JobVector*>(jobVector);
  size_type rows = self->size();

# if DEBUG_JOBLIST
  cerr << "timeoutCallback() collapseCountdown=" << self->collapseCountdown
       << endl;
  size_type realEntryCount = 0;
  for (size_type row = 0; row < rows; ++row)
    if (self->get(row) != 0) ++realEntryCount;
  Paranoid(realEntryCount == self->entryCount());
# endif

  bool collapsing = false;
  if (self->collapseCountdown > 0) {
    --self->collapseCountdown;
    collapsing = self->collapseCountdown == 0;
  }

  /* If there are no needTicks() Jobs, just a collapse operation in
     the future, we may not need to scan through the list at all. */
  if (self->needTicks == 1 && self->collapseCountdown > 0) return TRUE;

  /* Walk through list. If collapseCountdown == 0, remove >=0 rows at
     the end, move other rows. */
  bool prevEmpty = false; // Was previous row empty?
  self->freeze();
  //____________________

  for (size_type row = 0; row < rows; ++row) {
    Job* job = self->get(row);
    // Collapse operation
    if (collapsing && job != 0 && prevEmpty) {
      // Move row by one position towards top of list
      Paranoid(NR_OF_COLUMNS == 2);
      copyCell(self->list(), row, row - 1, COLUMN_PROGRESS);
      copyCell(self->list(), row, row - 1, COLUMN_OBJECT);
      self->set(row - 1, self->get(row));
      self->makeRowBlank(row);
      /* NB we do *not* set prevEmpty=true here - this causes the
         "stripes" of alternating empty/non-empty rows in the list.
         The stripe effect is the Right Thing, because it ensures that
         a click on a row always either selects the row the user
         intended to, or an empty one, *never* the following one! */
    }
    prevEmpty = job == 0;
    if (job != 0) {
      /* NB the Job can delete itself, so "job" becomes invalid after
         tick() is called. */
      // Yargh! This syntax took me 30 minutes to get right:
      if (job->needTicks()) (job->*(job->tick))();
    }
  }
  //____________________

  // If last row is empty, delete it
  if (rows > 0 && self->get(rows - 1) == 0)
    gtk_clist_remove(self->list(), rows - 1);

  self->thaw();
  Paranoid(self->collapseCountdown >= 0);

  if (collapsing) {
    // If all empty rows have been removed, maybe stop callback:
    if (self->entryCount() == self->size())
      self->unregisterTicks();
    else
      self->collapseCountdown = COLLAPSE_INTERVAL / TICK_INTERVAL;
  }
  if (self->needTicks > 0) return TRUE;
# if DEBUG_JOBLIST
  cerr << "timeoutCallback() unregistered\n" << endl;
# endif
  return FALSE;
}

//======================================================================

DownloadJob::DownloadJob(const string& uri, const string& dest)
  : download(uri, this), destination(dest), dataSizeVal(0) { }
//________________________________________

/* Note: We do not destroy the widgets, this will be done
   automatically once the row is removed from the GtkCList. */
DownloadJob::~DownloadJob() {
  // cerr << "~DownloadJob"<<endl;
}
//______________________________________________________________________

// Create widgets for a download Job
void DownloadJob::run() {
  Paranoid(getHandler() == 0);

  struct stat fileInfo;
  int statResult = stat(destination.c_str(), &fileInfo);

  // destination is present, but error accessing it
  if (statResult != 0 && errno != ENOENT) {
    string err = subst(_("Cannot access `%1': %2"), destination,
                       strerror(errno));
    errorBox(err);
    jobVector()->erase(row()); // Causes "delete this;"
    return;
  }

  /* We perform a regular download if the destination is either a
     filename (as opposed to a dir name), or the URI is not a jigdo
     file. */
  if ((statResult != 0 && errno == ENOENT) // destination does not exist
      || (statResult == 0 && !S_ISDIR(fileInfo.st_mode)) // exists as file
      || !compareEnd(download.uri(), ".jigdo")) { // URI end not ".jigdo"
    // Show URL as object name
    gtk_clist_set_text(list(), row(), JobVector::COLUMN_PROGRESS,
                       _("Waiting"));
    GtkPixmap* icon = pixmapForURI(download.uri());
    GdkPixmap* pixmap;
    GdkBitmap* mask;
    gtk_pixmap_get(icon, &pixmap, &mask);
    gtk_clist_set_pixtext(list(), row(), JobVector::COLUMN_OBJECT,
                          download.uri().c_str(), JobVector::PIXMAP_SPACING,
                          pixmap, mask);
    download.run();
  } else {
    errorBox("Processing of .jigdo files not yet implemented");
    jobVector()->erase(row()); // Causes "delete this;"
  }

  //xxx callRegularly(showProgress);
}
//______________________________________________________________________

namespace {

  // Print something like "9999", "9999k", "9999M", "99.9M"
  void printSize(string& s, uint64 size) {
    if (size < 10000) {
      append(s, size).append(_("B"));
      return;
    }
    if (size < 1024 * 100) {
      append(s, static_cast<double>(size) / 1024.0).append(_("kB"));
      return;
    }
    if (size < 1024 * 10000) {
      append(s, size / 1024).append(_("kB"));
      return;
    }
    if (size < 1048576 * 100) {
      append(s, static_cast<double>(size) / 1048576.0).append(_("MB"));
      return;
    }
    append(s, size / 1048576).append(_("MB"));
    return;
  }

}
//________________________________________

void DownloadJob::dataSize(uint64 n) { dataSizeVal = n; }
//________________________________________

// Show progress info
void DownloadJob::newData(const byte* /*data*/, size_t /*size*/,
                          uint64 totalSize) {
  GdkPixmap* progress = JobVector::progress(totalSize, dataSizeVal);
  GdkPixmap* oldProgress;
  gtk_clist_get_pixtext(list(), row(), JobVector::COLUMN_PROGRESS,
                        0, 0, &oldProgress, 0);
  // TODO: add bytes/sec, update progress at .5 sec intervals
  if (progress == oldProgress) return;

  string s;
  if (dataSizeVal <= totalSize) {
    dataSizeVal = totalSize;
  } else {
    append(s, 100.0 * totalSize / dataSizeVal).append(_("%, "));
  }
  printSize(s, totalSize);
  if (dataSizeVal > 0) {
    s += _(" of ");
    printSize(s, dataSizeVal);
  }
  gtk_clist_set_pixtext(list(), row(), JobVector::COLUMN_PROGRESS,
                        s.c_str(), JobVector::PIXMAP_SPACING, progress, 0);
}
//________________________________________

void DownloadJob::finished() {
  string size;
  printSize(size, dataSizeVal);
  string s = subst(_("Finished - fetched %1"), size);
  gtk_clist_set_text(list(), row(), JobVector::COLUMN_PROGRESS, s.c_str());
  callRegularlyLater(JobVector::FINISHED_WAIT, &DownloadJob::exit);
}
//________________________________________

void DownloadJob::info(const string& message) {
  gtk_clist_set_text(list(), row(), JobVector::COLUMN_PROGRESS,
                     message.c_str());
}
//________________________________________

void DownloadJob::error(const string& message) {
  gtk_clist_set_text(list(), row(), JobVector::COLUMN_PROGRESS,
                     message.c_str());
}
//________________________________________

void DownloadJob::exit() {
  // Delete ourself from the list
  jobVector()->erase(row());
}
